/****************************************************************************
 * This file is part of the program MusicRater.                             *
 * MusicRater is a program used to rate your music while listening to them  *
 * Copyright (C) 2005, 2006 Michael "Serpedon" Walz   (serpedon@gmx.de)     *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as published by     *
 * the Free Software Foundation; either version 2 of the License, or        *
 * (at your option) any later version.                                      *
 *                                                                          *
 * This program is distributed in the hope that it will be useful,          *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 * GNU General Public License for more details.                             *
 *                                                                          *
 * You should have received a copy of the GNU General Public License along  *
 * with this program; if not, write to the Free Softwares Foundation, Inc., *
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA                 *
 ****************************************************************************/
 
# include "Klasse.CreateFileHash.hpp"
# include "Klasse.Config.hpp"
# include <QFile>
# include <QTimer>

/*static*/ qint64 C_CreateFileHash::m_curAmountOfSyncProcesses = 0;
/*static*/ QMutex C_CreateFileHash::m_mutex_amountOfSyncProcesses;


C_CreateFileHash::C_CreateFileHash ( const QString & f_filename,
		/* <-- */				const bool f_selfDestoryAfterWork /*= false*/, 
		/* <-- */				const C_CreateFileHash::HashTypeFlag f_hashType /*= C_CreateFileHash::WholeFile*/ )
	: m_filename ( f_filename ), m_selfDestoryAfterWork ( f_selfDestoryAfterWork ), m_hashType ( f_hashType )
{
	QObject::connect ( & m_hashProcess, SIGNAL ( finished ( int ) ),
		/* <-- */		 this, SLOT ( ProcessFinished ( int ) ) );
}

void C_CreateFileHash::Start ( )
{
	bool delayStart = false;
	
	{
		QMutexLocker locker ( & m_mutex_amountOfSyncProcesses );
		
		if ( m_curAmountOfSyncProcesses >= m_maxAmountOfSyncProcesses )
			delayStart = true;
		else
			m_curAmountOfSyncProcesses++;
	}
	
	if ( delayStart )
		QTimer::singleShot ( 1000, this, SLOT ( Start ( ) ) );
	else
	{
		if ( G_config.GetProgGenerateHash ( ).isEmpty ( ) )
			ErrorHashingFailed ( HashCommandNotSet );
		else
		{
			
			m_hashProcess.start ( G_config.GetProgGenerateHash ( ) );
			
			QFile file ( m_filename );
			if ( ! file.open ( QIODevice::ReadOnly ) )
				ErrorHashingFailed ( CannotOpenFile );
			else
			{
				char buffer [ m_bufferSize ];
				
				// Buffersize muss mindestens 10 sein, damit der gesamten Header eines evtl.
				// 	vorhandenen MP3-ID3V2-Tag auf einmal gelesen werden kann
				// Buffersize muss mindestens 128 sein, damit der gesamten ID2-Tag am Ende einer Datei
				// 	auf einmal gelesen werden kann
				#define C_ASSERT(e) typedef char __C_ASSERT__[(e)?1:-1]
				C_ASSERT ( m_bufferSize >= 128 );
				
				if ( m_hashType & IgnoreFrontID3 )
				{
					// Es wird der ID3-Tag untersucht
					// vgl. http://www.id3.org/id3v2.4.0-structure.txt
					
					qint64 amountRead = file.read ( buffer, 10 );
					
					// berprfen, ob ein Fehler aufgetreten ist
					if ( amountRead < 0 )
					{
						ErrorHashingFailed ( CannotReadFromFile );
						return;
					}
					
					if (	buffer [ 0 ] == 'I' &&
							buffer [ 1 ] == 'D' &&
							buffer [ 2 ] == '3' )
					{
						bool hasFooter = buffer [ 5 ] & ( 0x1 << 4 );
						qint64 tagSize = buffer [ 9 ] 
											| buffer [ 8 ] << ( 7			)
											| buffer [ 7 ] << ( 7 + 8		)
											| buffer [ 6 ] << ( 7 + 8 + 8	);
											  
						// Die 10 Bytes des Header hinzufgen + des evtl. vorhanden Footers
						qint64 wholeHeaderSize = tagSize + 10 + ( hasFooter ? 10 : 0 );
						
						file.seek ( wholeHeaderSize );
						
						// Nach dem Sync-Byte suchen
						amountRead = file.read ( buffer, 1 );
						
						if ( amountRead != 1 || buffer [ 0 ] != ( char ) 0xFF )
						{
							ErrorHashingFailed ( ID3SyncByteNotFound );
							return;
						}
						else
						  file.seek ( wholeHeaderSize );
					}
					else
						file.seek ( 0 );
				}
				
				qint64 fileSize = file.size ( );
				
				while ( ! file.atEnd ( ) )
				{
					qint64 amountRead = file.read ( buffer, m_bufferSize );
					
					// berprfen, ob ein Fehler aufgetreten ist
					if ( amountRead < 0 )
					{
						ErrorHashingFailed ( CannotReadFromFile );
						return;
					}
					
					if ( m_hashType & IgnoreBackID2 )
					{
						qint64 filePos = file.pos ( );
						if ( filePos - amountRead == fileSize - m_sizeOfID2Tag )
						{
							if (	buffer [ 0 ] == 'T' &&
									buffer [ 1 ] == 'A' &&
									buffer [ 2 ] == 'G' )
							{
								// Aus der while-Schleife ausbrechen, da der TAG ignoriert werden soll
								break;
							}
						}
						else if ( filePos > fileSize - m_sizeOfID2Tag )
						{
							amountRead -= filePos - ( fileSize - m_sizeOfID2Tag );
							file.seek ( fileSize - m_sizeOfID2Tag );
						}
					}
					
					qint64 amountWritten = m_hashProcess.write ( buffer, amountRead );
					
					if ( amountRead != amountWritten )
					{
						ErrorHashingFailed ( CannotWriteToHashCommand );
						return;
					}			
				}
				
				m_hashProcess.closeWriteChannel ( );
			}
		}
	}
}

void C_CreateFileHash::ErrorHashingFailed ( C_CreateFileHash::FailedReasons f_failedReason )
{
	emit HashingFailed ( f_failedReason );
	SelfDestroy ( );
}


void C_CreateFileHash::ProcessFinished ( int f_exitCode )
{
	// ExitCode 0 bedeutet Erfolgreich
	if ( f_exitCode != 0 )
		ErrorHashingFailed ( ProcessDidNotSuccess );
	else
	{
		QString hash = ( ( QString ) m_hashProcess.readAllStandardOutput ( ) ).section ( ' ', 0, 0 );
		emit HashCode ( hash, m_filename, m_hashType );
		SelfDestroy ( );
	}
}

void C_CreateFileHash::SelfDestroy ( )
{
	{
		QMutexLocker locker ( & m_mutex_amountOfSyncProcesses );
		m_curAmountOfSyncProcesses--;
	}
	
	if ( m_selfDestoryAfterWork )
		delete this;
}
